package java.util.concurrent;

import java.util.function.Consumer;
import java.util.Spliterators;
import java.util.NoSuchElementException;
import java.util.Spliterator;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collection;
import sun.misc.Unsafe;
import java.io.Serializable;
import java.util.Queue;
import java.util.AbstractQueue;

public class ConcurrentLinkedQueue<E> extends AbstractQueue<E> implements Queue<E>, Serializable
{
    private static final long serialVersionUID = 196745693267521676L;
    private transient volatile ConcurrentLinkedQueue.Node<E> head;
    private transient volatile ConcurrentLinkedQueue.Node<E> tail;
    private static final Unsafe UNSAFE;
    private static final long headOffset;
    private static final long tailOffset;
    
    public ConcurrentLinkedQueue() {
        ConcurrentLinkedQueue.Node node = new ConcurrentLinkedQueue.Node(null);
        this.tail = node;
        this.head = node;
    }
    
    public ConcurrentLinkedQueue(Collection<? extends E> collection) {
        Object o;
        for (Object next : collection) {
            checkNotNull(next);
            o = new ConcurrentLinkedQueue.Node(next);
            if (o == null) {
                continue;
            }
            o.lazySetNext(o);
        }
        if (o == null) {
            ConcurrentLinkedQueue.Node node = new ConcurrentLinkedQueue.Node(null);
        }
        this.head = o;
        this.tail = o;
    }
    
    @Override
    public boolean add(E e) {
        return this.offer(e);
    }
    
    final void updateHead(ConcurrentLinkedQueue.Node<E> node, ConcurrentLinkedQueue.Node<E> node2) {
        if (node != node2 && this.casHead(node, node2)) {
            node.lazySetNext(node);
        }
    }
    
    final ConcurrentLinkedQueue.Node<E> succ(ConcurrentLinkedQueue.Node<E> node) {
        ConcurrentLinkedQueue.Node<Object> next = node.next;
        return (node == next) ? this.head : next;
    }
    
    @Override
    public boolean offer(E e) {
        checkNotNull(e);
        ConcurrentLinkedQueue.Node<E> node = new ConcurrentLinkedQueue.Node(e);
        ConcurrentLinkedQueue.Node<E> node2 = this.tail;
        while (true) {
            ConcurrentLinkedQueue.Node<Object> next = node2.next;
            if (next == null) {
                if (node2.casNext(null, node)) {
                    break;
                }
                continue;
            }
            else if (node2 == next) {
                node2 = ((node2 != (node2 = this.tail)) ? node2 : this.head);
            }
            else {
                node2 = ((node2 != node2 && node2 != (node2 = this.tail)) ? node2 : next);
            }
        }
        if (node2 != node2) {
            this.casTail(node2, node);
        }
        return true;
    }
    
    @Override
    public E poll() {
        while (true) {
            ConcurrentLinkedQueue.Node<E> head = this.head;
            while (true) {
                Object item = head.item;
                if (item != null && head.casItem(item, null)) {
                    if (head != head) {
                        ConcurrentLinkedQueue.Node<Object> next2;
                        this.updateHead(head, ((next2 = head.next) != null) ? next2 : head);
                        return item;
                    }
                    return item;
                }
                else {
                    ConcurrentLinkedQueue.Node<Object> next;
                    if ((next = head.next) == null) {
                        this.updateHead(head, head);
                        return null;
                    }
                    if (head == next) {
                        break;
                    }
                    head = next;
                }
            }
        }
    }
    
    @Override
    public E peek() {
        ConcurrentLinkedQueue.Node<E> head;
        Object item;
    Label_0027:
        while (true) {
            head = (head = this.head);
            while (true) {
                item = head.item;
                ConcurrentLinkedQueue.Node<Object> next;
                if (item != null || (next = head.next) == null) {
                    break Label_0027;
                }
                if (head == next) {
                    break;
                }
                head = next;
            }
        }
        this.updateHead(head, head);
        return item;
    }
    
    ConcurrentLinkedQueue.Node<E> first() {
        ConcurrentLinkedQueue.Node<E> head;
        int n;
    Label_0035:
        while (true) {
            head = (head = this.head);
            while (true) {
                n = (head.item != null);
                ConcurrentLinkedQueue.Node<Object> next;
                if (n || (next = head.next) == null) {
                    break Label_0035;
                }
                if (head == next) {
                    break;
                }
                head = next;
            }
        }
        this.updateHead(head, head);
        return n ? head : null;
    }
    
    @Override
    public boolean isEmpty() {
        return this.first() == null;
    }
    
    @Override
    public int size() {
        int n = 0;
        for (ConcurrentLinkedQueue.Node node = this.first(); node != null && (node.item == null || ++n != Integer.MAX_VALUE); node = this.succ(node)) {}
        return n;
    }
    
    @Override
    public boolean contains(Object o) {
        if (o == null) {
            return false;
        }
        ConcurrentLinkedQueue.Node<E> node = this.first();
        while (node != null) {
            Object item = node.item;
            if (item != null && o.equals(item)) {
                return true;
            }
            node = this.succ(node);
        }
        return false;
    }
    
    @Override
    public boolean remove(Object o) {
        if (o != null) {
            ConcurrentLinkedQueue.Node<Object> node = null;
            ConcurrentLinkedQueue.Node<E> node2;
            for (ConcurrentLinkedQueue.Node first = this.first(); first != null; node = first, first = node2) {
                boolean casItem = false;
                Object item = first.item;
                if (item != null) {
                    if (!o.equals(item)) {
                        node2 = this.succ(first);
                        continue;
                    }
                    casItem = first.casItem(item, null);
                }
                node2 = this.succ(first);
                if (node != null && node2 != null) {
                    node.casNext(first, node2);
                }
                if (!(casItem == 0)) {
                    return true;
                }
            }
        }
        return false;
    }
    
    @Override
    public boolean addAll(Collection<? extends E> collection) {
        if (collection == this) {
            throw new IllegalArgumentException();
        }
        Object o;
        for (Object next : collection) {
            checkNotNull(next);
            o = new ConcurrentLinkedQueue.Node(next);
            if (o == null) {
                continue;
            }
            o.lazySetNext(o);
        }
        if (o == null) {
            return false;
        }
        ConcurrentLinkedQueue.Node<E> node = this.tail;
        while (true) {
            ConcurrentLinkedQueue.Node<Object> next2 = node.next;
            if (next2 == null) {
                if (node.casNext(null, o)) {
                    break;
                }
                continue;
            }
            else if (node == next2) {
                node = ((node != (node = this.tail)) ? node : this.head);
            }
            else {
                node = ((node != node && node != (node = this.tail)) ? node : next2);
            }
        }
        if (!this.casTail(node, o)) {
            ConcurrentLinkedQueue.Node<E> tail = this.tail;
            if (o.next == null) {
                this.casTail(tail, o);
            }
        }
        return true;
    }
    
    @Override
    public Object[] toArray() {
        ArrayList<Object> list = new ArrayList();
        for (ConcurrentLinkedQueue.Node node = this.first(); node != null; node = this.succ(node)) {
            Object item = node.item;
            if (item != null) {
                list.add(item);
            }
        }
        return list.toArray();
    }
    
    @Override
    public <T> T[] toArray(T[] a) {
        int n;
        ConcurrentLinkedQueue.Node<E> node;
        Object item;
        for (n = 0, node = this.first(); node != null && n < a.length; node = this.succ(node)) {
            item = node.item;
            if (item != null) {
                a[n++] = item;
            }
        }
        if (node == null) {
            if (n < a.length) {
                a[n] = null;
                return a;
            }
            return a;
        }
        else {
            ArrayList o = new ArrayList();
            for (ConcurrentLinkedQueue.Node node2 = this.first(); node2 != null; node2 = this.succ(node2)) {
                Object item2 = node2.item;
                if (item2 != null) {
                    o.add(item2);
                }
            }
            return o.toArray(a);
        }
    }
    
    @Override
    public Iterator<E> iterator() {
        return new ConcurrentLinkedQueue.Itr();
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        objectOutputStream.defaultWriteObject();
        for (ConcurrentLinkedQueue.Node node = this.first(); node != null; node = this.succ(node)) {
            Object item = node.item;
            if (item != null) {
                objectOutputStream.writeObject(item);
            }
        }
        objectOutputStream.writeObject(null);
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        objectInputStream.defaultReadObject();
        Object object;
        Object o;
        while ((object = objectInputStream.readObject()) != null) {
            o = new ConcurrentLinkedQueue.Node(object);
            if (o == null) {
                continue;
            }
            o.lazySetNext(o);
        }
        if (o == null) {
            ConcurrentLinkedQueue.Node node = new ConcurrentLinkedQueue.Node(null);
        }
        this.head = o;
        this.tail = o;
    }
    
    @Override
    public Spliterator<E> spliterator() {
        return new ConcurrentLinkedQueue.CLQSpliterator(this);
    }
    
    private static void checkNotNull(Object o) {
        if (o == null) {
            throw new NullPointerException();
        }
        return;
    }
    
    private boolean casTail(ConcurrentLinkedQueue.Node<E> expected, ConcurrentLinkedQueue.Node<E> x) {
        return ConcurrentLinkedQueue.UNSAFE.compareAndSwapObject(this, ConcurrentLinkedQueue.tailOffset, expected, x);
    }
    
    private boolean casHead(ConcurrentLinkedQueue.Node<E> expected, ConcurrentLinkedQueue.Node<E> x) {
        return ConcurrentLinkedQueue.UNSAFE.compareAndSwapObject(this, ConcurrentLinkedQueue.headOffset, expected, x);
    }
    
    static {
        try {
            UNSAFE = Unsafe.getUnsafe();
            Class<Object> clazz = ConcurrentLinkedQueue.class;
            headOffset = ConcurrentLinkedQueue.UNSAFE.objectFieldOffset(clazz.getDeclaredField("head"));
            tailOffset = ConcurrentLinkedQueue.UNSAFE.objectFieldOffset(clazz.getDeclaredField("tail"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
    
    private static class Node<E>
    {
        volatile E item;
        volatile ConcurrentLinkedQueue.Node<E> next;
        private static final Unsafe UNSAFE;
        private static final long itemOffset;
        private static final long nextOffset;
        
        Node(E x) {
            ConcurrentLinkedQueue.Node.UNSAFE.putObject(this, ConcurrentLinkedQueue.Node.itemOffset, x);
        }
        
        boolean casItem(E expected, E x) {
            return ConcurrentLinkedQueue.Node.UNSAFE.compareAndSwapObject(this, ConcurrentLinkedQueue.Node.itemOffset, expected, x);
        }
        
        void lazySetNext(ConcurrentLinkedQueue.Node<E> x) {
            ConcurrentLinkedQueue.Node.UNSAFE.putOrderedObject(this, ConcurrentLinkedQueue.Node.nextOffset, x);
        }
        
        boolean casNext(ConcurrentLinkedQueue.Node<E> expected, ConcurrentLinkedQueue.Node<E> x) {
            return ConcurrentLinkedQueue.Node.UNSAFE.compareAndSwapObject(this, ConcurrentLinkedQueue.Node.nextOffset, expected, x);
        }
        
        static {
            try {
                UNSAFE = Unsafe.getUnsafe();
                Class<Object> clazz = ConcurrentLinkedQueue.Node.class;
                itemOffset = ConcurrentLinkedQueue.Node.UNSAFE.objectFieldOffset(clazz.getDeclaredField("item"));
                nextOffset = ConcurrentLinkedQueue.Node.UNSAFE.objectFieldOffset(clazz.getDeclaredField("next"));
            }
            catch (Exception cause) {
                throw new Error(cause);
            }
        }
    }
    
    private class Itr implements Iterator<E>
    {
        private ConcurrentLinkedQueue.Node<E> nextNode;
        private E nextItem;
        private ConcurrentLinkedQueue.Node<E> lastRet;
        
        Itr() {
            this.advance();
        }
        
        private E advance() {
            this.lastRet = this.nextNode;
            Object nextItem = this.nextItem;
            ConcurrentLinkedQueue.Node<Object> nextNode;
            ConcurrentLinkedQueue.Node<Object> nextNode2;
            if (this.nextNode == null) {
                nextNode2 = ConcurrentLinkedQueue.this.first();
                nextNode = null;
            }
            else {
                nextNode = this.nextNode;
                nextNode2 = ConcurrentLinkedQueue.this.succ(this.nextNode);
            }
            while (nextNode2 != null) {
                Object item = nextNode2.item;
                if (item != null) {
                    this.nextNode = nextNode2;
                    this.nextItem = item;
                    return nextItem;
                }
                ConcurrentLinkedQueue.Node<Object> succ = ConcurrentLinkedQueue.this.succ(nextNode2);
                if (nextNode != null && succ != null) {
                    nextNode.casNext(nextNode2, succ);
                }
                nextNode2 = succ;
            }
            this.nextNode = null;
            this.nextItem = null;
            return nextItem;
        }
        
        @Override
        public boolean hasNext() {
            return this.nextNode != null;
        }
        
        @Override
        public E next() {
            if (this.nextNode == null) {
                throw new NoSuchElementException();
            }
            return this.advance();
        }
        
        @Override
        public void remove() {
            ConcurrentLinkedQueue.Node<Object> lastRet = this.lastRet;
            if (lastRet == null) {
                throw new IllegalStateException();
            }
            lastRet.item = null;
            this.lastRet = null;
            return;
        }
    }
    
    static final class CLQSpliterator<E> implements Spliterator<E>
    {
        static final int MAX_BATCH = 33554432;
        final ConcurrentLinkedQueue<E> queue;
        ConcurrentLinkedQueue.Node<E> current;
        int batch;
        boolean exhausted;
        
        CLQSpliterator(ConcurrentLinkedQueue<E> queue) {
            this.queue = queue;
        }
        
        @Override
        public Spliterator<E> trySplit() {
            ConcurrentLinkedQueue<Object> queue = this.queue;
            int batch = this.batch;
            int n = (batch <= 0) ? 1 : ((batch >= 33554432) ? 33554432 : (batch + 1));
            ConcurrentLinkedQueue.Node<E> current;
            if (!this.exhausted && ((current = this.current) != null || (current = queue.first()) != null) && current.next != null) {
                Object[] array = new Object[n];
                int n2 = 0;
                do {
                    if ((array[n2] = current.item) != null) {
                        ++n2;
                    }
                    if (current == (current = current.next)) {
                        current = queue.first();
                    }
                    else {
                        continue;
                    }
                } while (current != null && n2 < n);
                if ((this.current = current) == null) {
                    this.exhausted = true;
                }
                if (n2 > 0) {
                    this.batch = n2;
                    return Spliterators.spliterator(array, 0, n2, 4368);
                }
            }
            return null;
        }
        
        @Override
        public void forEachRemaining(Consumer<? super E> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            ConcurrentLinkedQueue<Object> queue = this.queue;
            ConcurrentLinkedQueue.Node<E> node;
            if (!this.exhausted && ((node = this.current) != null || (node = queue.first()) != null)) {
                this.exhausted = true;
                do {
                    Object item = node.item;
                    if (node == (node = node.next)) {
                        node = queue.first();
                    }
                    if (item != null) {
                        consumer.accept(item);
                    }
                    else {
                        continue;
                    }
                } while (node != null);
            }
            return;
        }
        
        @Override
        public boolean tryAdvance(Consumer<? super E> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            ConcurrentLinkedQueue<Object> queue = this.queue;
            ConcurrentLinkedQueue.Node<E> current;
            if (!this.exhausted && ((current = this.current) != null || (current = queue.first()) != null)) {
                Object item;
                do {
                    item = current.item;
                    if (current == (current = current.next)) {
                        current = queue.first();
                    }
                    else {
                        continue;
                    }
                } while (item == null && current != null);
                if ((this.current = current) == null) {
                    this.exhausted = true;
                }
                if (item != null) {
                    consumer.accept(item);
                    return true;
                }
            }
            return false;
        }
        
        @Override
        public long estimateSize() {
            return Long.MAX_VALUE;
        }
        
        @Override
        public int characteristics() {
            return 4368;
        }
    }
}
